#!/bin/bash

usage() {
	echo -e "\e[01;31m`basename $0`\e[00m [-h | --help] [-t | --test] [--baseonly] [--full] [--without-boost] [--without-debug] [[-c | --config] config]"
	echo -e "    \e[01;34m--h | --help\e[00m\tShow This Help"
	echo -e "    \e[01;34m--t | --test\e[00m\tBuild Kernel With '.fortest' Lable"
	echo -e "    \e[01;34m--baseonly\e[00m\t\tBuild Base Kernel Only, Without Boost and Debug Version"
	echo -e "    \e[01;34m--boostonly\e[00m\t\tBuild Performance Boost Version Only, Without Standard and Debug Version"
	echo -e "    \e[01;34m--debugonly\e[00m\t\tBuild Debug Kernel Only, Without Standard and Boost version"
	echo -e "    \e[01;34m--full\e[00m\t\tBuild Full Kernel"
	echo -e "    \e[01;34m--without-boost\e[00m\tBuild Full Kernel Without Performance Boost Version"
	echo -e "    \e[01;34m--without-debug\e[00m\tBuild Full Kernel Without Debug Version"
	echo -e "    \e[01;34m--without-up\e[00m\tBuild Full Kernel Without Standard Version"
	echo -e "    \e[01;34m-c | --config\e[00m\tBuild Test Kernel With Specify Config"
	echo -e "    \e[01;34m-l | --label\e[00m\tBuild Kernel With Specify Lable"
}

if [ $# -lt 1 ]; then
	usage
	exit -1;
fi

BUILDUP=true
BUILDBOOST=true
BUILDDEBUG=true
BUILDLAB=false
BUILDTEST=false
TESTLAB=""

ARGS=`getopt -o htc:l: -a --long help,test,baseonly,boostonly,debugonly,full,without-boost,without-debug,without-up,config,label -- "$@"`
if [ $? != 0 ]; then
	echo "Terminating..."
	usage && exit 1
fi

eval set -- "${ARGS}"

while true
do
	case $1 in
		--baseonly)
			BUILDBOOST=false
			BUILDDEBUG=false
			shift;;
		--boostonly)
			BUILDUP=false
			BUILDDEBUG=false
			shift;;
		--debugonly)
			BUILDBOOST=false
			BUILDUP=false
			shift;;
		--full)
			shift;;
		--without-boost)
			BUILDBOOST=false
			shift;;
		--without-debug)
			BUILDDEBUG=false
			shift;;
		--without-up)
			BUILDUP=false
			shift;;
		-t|--test)
			BUILDTEST=true
			TESTLAB=".fortest"
			shift;;
		-c|--config)
			case "$2" in
				"")
					echo "Option c, no argument";
					exit -1;;
				*)
					BUILDBOOST=false
					BUILDDEBUG=false
					Config=$2
					shift 2;;
			esac;;
		-l|--label)
			case "$2" in
				"")
					echo "Option l, no argument";
					exit -1;;
				*)
					BUILDLAB=true
					TESTLAB=".$2"
					shift 2;;
			esac;;
		--)
			shift
			break;;
		-h|--help|*)
			usage
			exit -1;
	esac
done

for arg in $@
do
	echo "Ignore Unknow Args $arg.."
	usage && exit -1
done

ARCH=`uname -m`
Arch=`echo $ARCH | sed -e s/i.86/x86/ -e s/x86_64/x86/ -e s/amd64/x86/ -e s/aarch64.*/arm64/ -e s/mips64el/mips/ -e s/loongarch64.*/loongarch/`
TOPDIR=`pwd`/rpmbuild

if [ "$Config" != "" ]; then
	CLEAN="-e $Config"
	BUILDTEST=true
	\cp $Config arch/${Arch}/configs/kylinos_generic_defconfig
	if [ $? != 0 ]; then
		exit -1;
	fi
fi

git clean -xdf -e nohup.out $CLEAN
rm -rf rpmbuild/BUILD
rm -rf rpmbuild/BUILDROOT

rm -rf rpmbuild/linux-5.10.0/
mkdir rpmbuild/linux-5.10.0/
make mrproper -j 16

rsync -av --exclude rpmbuild --exclude .git . rpmbuild/linux-5.10.0/

#prepare linux-5.10.0.tar.gz
cd rpmbuild

tar -zcvf linux-5.10.0.tar.xz linux-5.10.0
cd ../
rm -rf rpmbuild/SOURCES/linux-5.10.0.tar.xz
mv rpmbuild/linux-5.10.0.tar.xz rpmbuild/SOURCES

#do building
BUILD_ARGS=""

if [[ $BUILDBOOST == false ]]; then
	BUILD_ARGS="$BUILD_ARGS --without boost"
fi
if [[ $BUILDDEBUG == false ]]; then
	BUILD_ARGS="$BUILD_ARGS --without debug"
fi
if [[ $BUILDUP == false ]]; then
	BUILD_ARGS="$BUILD_ARGS --without up"
fi

if [[ $BUILDTEST == true ]] || [[ $BUILDLAB == true ]]; then
	rpmbuild -bb rpmbuild/SPECS/kernel.spec --define="_topdir $TOPDIR" ${BUILD_ARGS} --define="custom ${TESTLAB}"
else
	rpmbuild -bb rpmbuild/SPECS/kernel.spec --define="_topdir $TOPDIR" ${BUILD_ARGS}
fi

#clean work
if [ $? -eq 0 ]; then
	echo "building rpm success..."
#	rm -rf rpmbuild/kernel
else
	echo "building failed..."
	exit
fi

#archive RPMS
MIDV=`grep -rn "release_major" rpmbuild/SPECS/kernel.spec  | head  -n 1  | awk -F' ' '{print $3}'`
SUBV=`grep -rn "release_minor" rpmbuild/SPECS/kernel.spec  | head  -n 1  | awk -F' ' '{print $3}'`
IFRELEASE=`grep -rn "global released_kernel" rpmbuild/SPECS/kernel.spec  | head  -n 1  | awk -F' ' '{print $3}'`
RPMDIR=${TOPDIR}/RPMS/${ARCH}
if [ $IFRELEASE -eq 1 ];then
	KVRELEASE=kernel-5.10.0-${MIDV}${SUBV}${TESTLAB}-${ARCH}
else
	KVRELEASE=kernel-5.10.0-${MIDV}${TESTLAB}-${ARCH}
fi
rm -rf rpmbuild/$KVRELEASE*
mkdir -p rpmbuild/$KVRELEASE

for pack in `ls $RPMDIR/${rpmpkg}*.rpm | grep -v debuginfo-common`
do
	cp $pack rpmbuild/$KVRELEASE
done

if [[ $BUILDBOOST != false ]]; then
	mkdir -p rpmbuild/$KVRELEASE/boost-version
	mv rpmbuild/$KVRELEASE/*boost*.rpm rpmbuild/$KVRELEASE/boost-version
	VERSIONS="boost-version"
fi

if [[ $BUILDDEBUG != false ]]; then
	mkdir -p rpmbuild/$KVRELEASE/debug-version
	mv rpmbuild/$KVRELEASE/*kernel-debug-*.rpm rpmbuild/$KVRELEASE/debug-version
	VERSIONS="$VERSIONS debug-version"
fi

mkdir -p rpmbuild/$KVRELEASE/all
mv rpmbuild/$KVRELEASE/{*tool*,*perf*,kernel-cross-headers*} rpmbuild/$KVRELEASE/all

if [[ $BUILDUP != false ]]; then
	mkdir -p rpmbuild/$KVRELEASE/base-version
	mv rpmbuild/$KVRELEASE/*.rpm rpmbuild/$KVRELEASE/base-version
	VERSIONS="$VERSIONS base-version"
fi

for version in $VERSIONS
do
	cp rpmbuild/$KVRELEASE/all/* rpmbuild/$KVRELEASE/$version
	echo "This is a $version kernel package: 5.10.0-${MIDV}.${SUBV}-${ARCH}-${version}" > rpmbuild/$KVRELEASE/$version/README
	cd rpmbuild/$KVRELEASE/
	tar -zcvf ${version}.tar.gz $version
	rm -rf $version
	cd -
done

rm -rf rpmbuild/$KVRELEASE/all

#clean
cd rpmbuild
tar -zcvf ${KVRELEASE}.tar.gz $KVRELEASE
rm -rf $KVRELEASE
echo -e "Archive all builded rpm packages in: \nrpmbuild/${KVRELEASE}.tar.gz"
cd ..
